@@ -1,6 +1,8 @@
 from __future__ import annotations
 
-from typing import TYPE_CHECKING, Any, Callable, Literal, Protocol, cast
+# ruff: noqa: N806
+import operator as op
+from typing import TYPE_CHECKING, Any, Callable, Literal, Protocol
 
 from narwhals._compliant.expr import LazyExpr
 from narwhals._compliant.typing import (
@@ -25,7 +27,10 @@ if TYPE_CHECKING:
 
     from narwhals._compliant.typing import AliasNames, WindowFunction
     from narwhals._expression_parsing import ExprMetadata
-    from narwhals.typing import PythonLiteral, RankMethod
+    from narwhals._sql.expr_dt import SQLExprDateTimeNamesSpace
+    from narwhals._sql.expr_str import SQLExprStringNamespace
+    from narwhals._sql.namespace import SQLNamespace
+    from narwhals.typing import NumericLiteral, PythonLiteral, RankMethod, TemporalLiteral
 
 
 class SQLExpr(LazyExpr[SQLLazyFrameT, NativeExprT], Protocol[SQLLazyFrameT, NativeExprT]):
@@ -51,6 +56,10 @@ class SQLExpr(LazyExpr[SQLLazyFrameT, NativeExprT], Protocol[SQLLazyFrameT, Nati
     def __call__(self, df: SQLLazyFrameT) -> Sequence[NativeExprT]:
         return self._call(df)
 
+    def __narwhals_namespace__(
+        self,
+    ) -> SQLNamespace[SQLLazyFrameT, Self, Any, NativeExprT]: ...
+
     def _callable_to_eval_series(
         self, call: Callable[..., NativeExprT], /, **expressifiable_args: Self | Any
     ) -> EvalSeries[SQLLazyFrameT, NativeExprT]:
@@ -140,11 +149,19 @@ class SQLExpr(LazyExpr[SQLLazyFrameT, NativeExprT], Protocol[SQLLazyFrameT, Nati
         )
 
     def _with_alias_output_names(self, func: AliasNames | None, /) -> Self:
+        current_alias_output_names = self._alias_output_names
+        alias_output_names = (
+            None
+            if func is None
+            else func
+            if current_alias_output_names is None
+            else lambda output_names: func(current_alias_output_names(output_names))
+        )
         return type(self)(
             self._call,
             self._window_function,
             evaluate_output_names=self._evaluate_output_names,
-            alias_output_names=func,
+            alias_output_names=alias_output_names,
             version=self._version,
             implementation=self._implementation,
         )
@@ -162,10 +179,25 @@ class SQLExpr(LazyExpr[SQLLazyFrameT, NativeExprT], Protocol[SQLLazyFrameT, Nati
 
         return self._window_function or default_window_func
 
-    def _function(self, name: str, *args: NativeExprT | PythonLiteral) -> NativeExprT: ...
-    def _lit(self, value: Any) -> NativeExprT: ...
+    def _function(self, name: str, *args: NativeExprT | PythonLiteral) -> NativeExprT:
+        return self.__narwhals_namespace__()._function(name, *args)
+
+    def _lit(self, value: Any) -> NativeExprT:
+        return self.__narwhals_namespace__()._lit(value)
+
+    def _coalesce(self, *expr: NativeExprT) -> NativeExprT:
+        return self.__narwhals_namespace__()._coalesce(*expr)
+
     def _count_star(self) -> NativeExprT: ...
-    def _when(self, condition: NativeExprT, value: NativeExprT) -> NativeExprT: ...
+
+    def _when(
+        self,
+        condition: NativeExprT,
+        value: NativeExprT,
+        otherwise: NativeExprT | None = None,
+    ) -> NativeExprT:
+        return self.__narwhals_namespace__()._when(condition, value, otherwise)
+
     def _window_expression(
         self,
         expr: NativeExprT,
@@ -247,7 +279,7 @@ class SQLExpr(LazyExpr[SQLLazyFrameT, NativeExprT], Protocol[SQLLazyFrameT, Nati
             }
             return [
                 self._when(
-                    self._window_expression(  # type: ignore[operator]
+                    self._window_expression(
                         self._function("count", expr), **window_kwargs
                     )
                     >= self._lit(min_samples),
@@ -359,6 +391,44 @@ class SQLExpr(LazyExpr[SQLLazyFrameT, NativeExprT], Protocol[SQLLazyFrameT, Nati
         return self._with_binary(lambda expr, other: expr.__or__(other), other)
 
     # Aggregations
+    def all(self) -> Self:
+        def f(expr: NativeExprT) -> NativeExprT:
+            return self._coalesce(self._function("bool_and", expr), self._lit(True))  # noqa: FBT003
+
+        def window_f(
+            df: SQLLazyFrameT, inputs: WindowInputs[NativeExprT]
+        ) -> Sequence[NativeExprT]:
+            return [
+                self._coalesce(
+                    self._window_expression(
+                        self._function("bool_and", expr), inputs.partition_by
+                    ),
+                    self._lit(True),  # noqa: FBT003
+                )
+                for expr in self(df)
+            ]
+
+        return self._with_callable(f)._with_window_function(window_f)
+
+    def any(self) -> Self:
+        def f(expr: NativeExprT) -> NativeExprT:
+            return self._coalesce(self._function("bool_or", expr), self._lit(False))  # noqa: FBT003
+
+        def window_f(
+            df: SQLLazyFrameT, inputs: WindowInputs[NativeExprT]
+        ) -> Sequence[NativeExprT]:
+            return [
+                self._coalesce(
+                    self._window_expression(
+                        self._function("bool_or", expr), inputs.partition_by
+                    ),
+                    self._lit(False),  # noqa: FBT003
+                )
+                for expr in self(df)
+            ]
+
+        return self._with_callable(f)._with_window_function(window_f)
+
     def max(self) -> Self:
         return self._with_callable(lambda expr: self._function("max", expr))
 
@@ -371,10 +441,58 @@ class SQLExpr(LazyExpr[SQLLazyFrameT, NativeExprT], Protocol[SQLLazyFrameT, Nati
     def min(self) -> Self:
         return self._with_callable(lambda expr: self._function("min", expr))
 
+    def count(self) -> Self:
+        return self._with_callable(lambda expr: self._function("count", expr))
+
+    def sum(self) -> Self:
+        def f(expr: NativeExprT) -> NativeExprT:
+            return self._coalesce(self._function("sum", expr), self._lit(0))
+
+        def window_f(
+            df: SQLLazyFrameT, inputs: WindowInputs[NativeExprT]
+        ) -> Sequence[NativeExprT]:
+            return [
+                self._coalesce(
+                    self._window_expression(
+                        self._function("sum", expr), inputs.partition_by
+                    ),
+                    self._lit(0),
+                )
+                for expr in self(df)
+            ]
+
+        return self._with_callable(f)._with_window_function(window_f)
+
     # Elementwise
     def abs(self) -> Self:
         return self._with_elementwise(lambda expr: self._function("abs", expr))
 
+    def clip(
+        self,
+        lower_bound: Self | NumericLiteral | TemporalLiteral | None,
+        upper_bound: Self | NumericLiteral | TemporalLiteral | None,
+    ) -> Self:
+        def _clip_lower(expr: NativeExprT, lower_bound: Any) -> NativeExprT:
+            return self._function("greatest", expr, lower_bound)
+
+        def _clip_upper(expr: NativeExprT, upper_bound: Any) -> NativeExprT:
+            return self._function("least", expr, upper_bound)
+
+        def _clip_both(
+            expr: NativeExprT, lower_bound: Any, upper_bound: Any
+        ) -> NativeExprT:
+            return self._function(
+                "greatest", self._function("least", expr, upper_bound), lower_bound
+            )
+
+        if lower_bound is None:
+            return self._with_elementwise(_clip_upper, upper_bound=upper_bound)
+        if upper_bound is None:
+            return self._with_elementwise(_clip_lower, lower_bound=lower_bound)
+        return self._with_elementwise(
+            _clip_both, lower_bound=lower_bound, upper_bound=upper_bound
+        )
+
     def is_null(self) -> Self:
         return self._with_elementwise(lambda expr: self._function("isnull", expr))
 
@@ -383,9 +501,32 @@ class SQLExpr(LazyExpr[SQLLazyFrameT, NativeExprT], Protocol[SQLLazyFrameT, Nati
             lambda expr: self._function("round", expr, self._lit(decimals))
         )
 
+    def sqrt(self) -> Self:
+        def _sqrt(expr: NativeExprT) -> NativeExprT:
+            return self._when(
+                expr < self._lit(0), self._lit(float("nan")), self._function("sqrt", expr)
+            )
+
+        return self._with_elementwise(_sqrt)
+
     def exp(self) -> Self:
         return self._with_elementwise(lambda expr: self._function("exp", expr))
 
+    def log(self, base: float) -> Self:
+        def _log(expr: NativeExprT) -> NativeExprT:
+            F = self._function
+            return self._when(
+                expr < self._lit(0),
+                self._lit(float("nan")),
+                self._when(
+                    expr == self._lit(0),
+                    self._lit(float("-inf")),
+                    op.truediv(F("log", expr), F("log", self._lit(base))),
+                ),
+            )
+
+        return self._with_elementwise(_log)
+
     # Cumulative
     def cum_sum(self, *, reverse: bool) -> Self:
         return self._with_window_function(self._cum_window_func("sum", reverse=reverse))
@@ -438,11 +579,10 @@ class SQLExpr(LazyExpr[SQLLazyFrameT, NativeExprT], Protocol[SQLLazyFrameT, Nati
         def func(
             df: SQLLazyFrameT, inputs: WindowInputs[NativeExprT]
         ) -> Sequence[NativeExprT]:
+            F = self._function
+            window = self._window_expression
             return [
-                expr  # type: ignore[operator]
-                - self._window_expression(
-                    self._function("lag", expr), inputs.partition_by, inputs.order_by
-                )
+                op.sub(expr, window(F("lag", expr), inputs.partition_by, inputs.order_by))
                 for expr in self(df)
             ]
 
@@ -467,15 +607,12 @@ class SQLExpr(LazyExpr[SQLLazyFrameT, NativeExprT], Protocol[SQLLazyFrameT, Nati
         ) -> Sequence[NativeExprT]:
             # pyright checkers think the return type is `list[bool]` because of `==`
             return [
-                cast(
-                    "NativeExprT",
-                    self._window_expression(
-                        self._function("row_number"),
-                        (*inputs.partition_by, expr),
-                        inputs.order_by,
-                    )
-                    == self._lit(1),
+                self._window_expression(
+                    self._function("row_number"),
+                    (*inputs.partition_by, expr),
+                    inputs.order_by,
                 )
+                == self._lit(1)
                 for expr in self(df)
             ]
 
@@ -486,17 +623,14 @@ class SQLExpr(LazyExpr[SQLLazyFrameT, NativeExprT], Protocol[SQLLazyFrameT, Nati
             df: SQLLazyFrameT, inputs: WindowInputs[NativeExprT]
         ) -> Sequence[NativeExprT]:
             return [
-                cast(
-                    "NativeExprT",
-                    self._window_expression(
-                        self._function("row_number"),
-                        (*inputs.partition_by, expr),
-                        inputs.order_by,
-                        descending=[True] * len(inputs.order_by),
-                        nulls_last=[True] * len(inputs.order_by),
-                    )
-                    == self._lit(1),
+                self._window_expression(
+                    self._function("row_number"),
+                    (*inputs.partition_by, expr),
+                    inputs.order_by,
+                    descending=[True] * len(inputs.order_by),
+                    nulls_last=[True] * len(inputs.order_by),
                 )
+                == self._lit(1)
                 for expr in self(df)
             ]
 
@@ -526,20 +660,27 @@ class SQLExpr(LazyExpr[SQLLazyFrameT, NativeExprT], Protocol[SQLLazyFrameT, Nati
                 "nulls_last": nulls_last,
             }
             count_window_kwargs: dict[str, Any] = {"partition_by": (*partition_by, expr)}
+            window = self._window_expression
+            F = self._function
             if method == "max":
-                rank_expr = (
-                    self._window_expression(func, **window_kwargs)  # type: ignore[operator]
-                    + self._window_expression(count_expr, **count_window_kwargs)
-                    - self._lit(1)
+                rank_expr = op.sub(
+                    op.add(
+                        window(func, **window_kwargs),
+                        window(count_expr, **count_window_kwargs),
+                    ),
+                    self._lit(1),
                 )
             elif method == "average":
-                rank_expr = self._window_expression(func, **window_kwargs) + (
-                    self._window_expression(count_expr, **count_window_kwargs)  # type: ignore[operator]
-                    - self._lit(1)
-                ) / self._lit(2.0)
+                rank_expr = op.add(
+                    window(func, **window_kwargs),
+                    op.truediv(
+                        op.sub(window(count_expr, **count_window_kwargs), self._lit(1)),
+                        self._lit(2.0),
+                    ),
+                )
             else:
-                rank_expr = self._window_expression(func, **window_kwargs)
-            return self._when(~self._function("isnull", expr), rank_expr)  # type: ignore[operator]
+                rank_expr = window(func, **window_kwargs)
+            return self._when(~F("isnull", expr), rank_expr)  # type: ignore[operator]
 
         def _unpartitioned_rank(expr: NativeExprT) -> NativeExprT:
             return _rank(expr, descending=[descending], nulls_last=[True])
@@ -568,11 +709,9 @@ class SQLExpr(LazyExpr[SQLLazyFrameT, NativeExprT], Protocol[SQLLazyFrameT, Nati
         def _is_unique(
             expr: NativeExprT, *partition_by: str | NativeExprT
         ) -> NativeExprT:
-            return cast(
-                "NativeExprT",
-                self._window_expression(self._count_star(), (expr, *partition_by))
-                == self._lit(1),
-            )
+            return self._window_expression(
+                self._count_star(), (expr, *partition_by)
+            ) == self._lit(1)
 
         def _unpartitioned_is_unique(expr: NativeExprT) -> NativeExprT:
             return _is_unique(expr)
@@ -602,9 +741,19 @@ class SQLExpr(LazyExpr[SQLLazyFrameT, NativeExprT], Protocol[SQLLazyFrameT, Nati
             implementation=self._implementation,
         )
 
+    # Namespaces
+    @property
+    def str(self) -> SQLExprStringNamespace[Self]: ...
+
+    @property
+    def dt(self) -> SQLExprDateTimeNamesSpace[Self]: ...
+
+    # Not implemented
+
     arg_max: not_implemented = not_implemented()
     arg_min: not_implemented = not_implemented()
     arg_true: not_implemented = not_implemented()
+    cat: not_implemented = not_implemented()  # type: ignore[assignment]
     drop_nulls: not_implemented = not_implemented()
     ewm_mean: not_implemented = not_implemented()
     gather_every: not_implemented = not_implemented()
@@ -616,6 +765,3 @@ class SQLExpr(LazyExpr[SQLLazyFrameT, NativeExprT], Protocol[SQLLazyFrameT, Nati
     tail: not_implemented = not_implemented()
     sample: not_implemented = not_implemented()
     unique: not_implemented = not_implemented()
-
-    # namespaces
-    cat: not_implemented = not_implemented()  # type: ignore[assignment]
